--constants
nodeinfosize = 336
objectinfosize = 192
meshinfosize = 208

struct nodeinfo   (offset,id,index,parentoffset,neighbouroffset,firstchildoffset,dataoffset,rawmatrix,bboxmin,bboxmax,inscene)
--objectinfo aka groupheader
struct objectinfo (offset,id,neighbouroffset,dataoffset,flags,bboxmin,bboxmax)
struct meshinfo (offset,subcount,materialcount,verticescount,indexbuffershorts,materialsoffset,vertbufferheader,indexbufferheader,secondsubcount,meshmaterials,beforevertbufferheader,elements)
struct materialinfo (offsetA,flags,offsetB,specA,specB)
struct meshelement (firstface,facecount,lowvertex,highvert)

---------------------------------------------------------------------------------------------------
function skip f bytes = 
	(
		here = ftell f
		fseek f (here + bytes) #seek_set
	)

function readpoint3 f pointref = 
	(
		for k = 1 to 3 do
		( 
			pointref[k] = (readfloat f)
		)						
	) 

function readquat f  = 
	(
		aquat = quat 0 0 0 1
		zero = readfloat f
		aquat.x = readfloat f
		aquat.y = readfloat f
		aquat.z = readfloat f
		aquat.w = readfloat f			
		return aquat
	)

function readextmatrix f  = 
	(			
		here = ftell f
		--format "	matrixdata : %\n" here				
		amatrix = matrix3 1
		quats = #()
		
		for k =  1 to 4 do
		(			
			aquat =  readquat f
			--print aquat
			append quats aquat
		)
		
		apoint = point3 quats[4].x quats[4].z quats[4].y			
		amatrix[4] = apoint					 
		return amatrix
	) 

function readnodeinfo f offset = 
	(			
		verbose = true				
		anodeinfo = nodeinfo 0 0 0 0 0 0 0 0 0					
		fseek f offset #seek_set
		--format "% : nodeinfo\n" offset
		anodeinfo.offset = offset					
		skip f (12*4)	
		
		here = ftell f
		anodeinfo.rawmatrix = matrix3 0				
		anodeinfo.rawmatrix = readextmatrix f
		--format "	% -> matrix % \n" 	here anodeinfo.rawmatrix	
		
		here = ftell f
		anodeinfo.dataoffset =  readlong f	
		--format "	[%] = % : dtaoffset\n" here anodeinfo.dataoffset
		
		here = ftell f
		anodeinfo.parentoffset = readlong f	
		--format "	[%] = % : parentoffset\n" here anodeinfo.parentoffset
		
		here = ftell f
		anodeinfo.firstchildoffset = readlong f
		--format "	[%] = % : firstchildoffset\n" here anodeinfo.firstchildoffset
		
		here = ftell f
		anodeinfo.neighbouroffset = readlong f						
		--format "	[%] = % : neighbouroffset\n" here anodeinfo.neighbouroffset
		
		skip f (32*4)					
		anodeinfo.bboxmin = point3 0 0 0 
		readpoint3 f anodeinfo.bboxmin			
		skip f 4			
		anodeinfo.bboxmax  = point3 0 0 0
		readpoint3 f	anodeinfo.bboxmax											
		skip f (5*4)	

		here = ftell f
		anodeinfo.index = readlong f
		--format "	[%] = % : index\n" here anodeinfo.index
		
		here = ftell f
		anodeinfo.id = readlong f
		--format "	[%] = % : id\n" here anodeinfo.id
		
		skip f (2*4)
		anodeinfo.inscene = false

		return anodeinfo																							
	)

function readobjectinfo f offset = 
	(						
		aobjectinfo = objectinfo 0 0 0 0 0 0 0
		fseek f offset #seek_set
		--format "% : objectinfo\n" offset
		aobjectinfo.offset = offset					
		aobjectinfo.neighbouroffset = readlong f					
		aobjectinfo.dataoffset = readlong f 						
		aobjectinfo.ID =  readlong f 	
		aobjectinfo.flags = readlong f
		skip f (32*4)	
		aobjectinfo.bboxmin = point3 0 0 0 
		readpoint3 f aobjectinfo.bboxmin					
		skip f 4
		aobjectinfo.bboxmax = point3 0 0 0 
		readpoint3 f aobjectinfo.bboxmax	
		return aobjectinfo								
	)

function readmaterialinfo f offset = 
	(
		amaterialinfo = materialinfo 0 0 0 0 0
		fseek f offset #seek_set
		--format "% : materialinfo\n" offset
		amaterialinfo.offsetA = readlong f
		amaterialinfo.flags	= readlong f
		
		here = ftell f
		amaterialinfo.offsetB	= readlong f
		--format "[%] = %  splitinfo\n" here amaterialinfo.offsetB
		
		amaterialinfo.specA	= readlong f
		amaterialinfo.specB	= readlong f
		return amaterialinfo
	)

function readmeshinfo f offset = 
	(
		ameshinfo = meshinfo 0 0 0 0 0 0 0 0 0
		fseek f offset #seek_set
		--format "% : meshinfo\n" offset
		skip f (2*4)
		ameshinfo.offset = offset		
		ameshinfo.subcount = readlong f
		
		here  = ftell f							
		ameshinfo.materialsoffset = readlong f			
		--format "[%] = %  materialsoffset\n" here ameshinfo.materialsoffset
		
		ameshinfo.beforevertbufferheader = readlong f					
		skip f (3*4)					
		ameshinfo.vertbufferheader = readlong f					
		ameshinfo.indexbufferheader = readlong f			
		ameshinfo.indexbuffershorts = readlong f
		skip f (36*4)	
		ameshinfo.secondsubcount = readlong f			
		ameshinfo.materialcount = readlong f
		ameshinfo.verticescount = readlong f
		here = ftell f
		--print here
		ameshinfo.meshmaterials = #()	
		for m = 0 to (ameshinfo.materialcount-1) do
		(
			amaterialinfo = readmaterialinfo f (m*5*4+ameshinfo.materialsoffset)
			append ameshinfo.meshmaterials amaterialinfo			
		)	
		
		ameshinfo.elements = #()
		for amatinf in ameshinfo.meshmaterials do
			(	
				fseek f amatinf.offsetB #seek_set
				--format "% : meshsplitinfo\n" amatinf.offsetB
				ameshelement = meshelement 0 0 0 0
				ameshelement.firstface = readlong f
				ameshelement.facecount = readlong f
				ameshelement.lowvertex = readlong f
				ameshelement.highvert  = readlong f
				--format "%\n" ameshelement
				append ameshinfo.elements ameshelement			
			)			
		return ameshinfo		
	)	

function readparentmatrix f offset matrixref  = 
	(		
		fseek f offset #seek_set	
		--format "% : matrix" offset
		skip f (12*4)											
		for k = 1 to 3 do
			(
				zero = readfloat f
				apoint = point3 0 0 0
				readpoint3 f apoint
				matrixref[k] = apoint
				one = readfloat f
			)
		zero = readfloat f
		apoint = point3 0 0 0
		readpoint3 f apoint
		asecpoint = point3 apoint[1] apoint[3] apoint[2]
		matrixref[4] = asecpoint
		one = readfloat f					 
	) 

function readvertexbuffer f offset count vertarrayref normalarrayref uvwarrayref = 
	(		
		fseek f offset #seek_set
		--format "% : vertexbuffer\n" offset
		for vi = 1 to count do
			(
				apoint = point3 0 0 0
				readpoint3 f apoint
				asecpoint = point3 apoint[1] apoint[3] apoint[2]
				append vertarrayref asecpoint
						
				apoint = point3 0 0 0
				readpoint3 f apoint
				apoint[1] = apoint[1] 
				append normalarrayref apoint 
						
				color = readfloat f
				
				apoint = point3 0 0 0
				apoint[1] = readfloat f	--u	
				apoint[2] = 1 - (readfloat f)	--v						
				append uvwarrayref apoint																
			)--for v = 1 to vertex_count					
	)

function readindexbuffer f offset count facearrayref  = 
	(		
		fseek f offset #seek_set					
		--format "% : indexbuffer\n" offset
		for fi = 1 to count  do
		(
			fd = readshort f
			fd += 1	
			append facearrayref fd
		)--for fi = 1 to (face_data_size/2) 					
	)		

function drawnode f anodeinfo = 
	(
		--print anodeinfo						
		aobjectinfo = readobjectinfo f (anodeinfo.offset + nodeinfosize)					
		--print aobjectinfo					
		ameshinfo = readmeshinfo f (aobjectinfo.offset + objectinfosize)
		--print ameshinfo

		vertex_array =#()
		normal_array = #()
		uvarray = #()
		readvertexbuffer f (ameshinfo.vertbufferheader + 3*4) ameshinfo.verticescount vertex_array normal_array uvarray
			
		rawindices_array = #()
		readindexbuffer f (ameshinfo.indexbufferheader + 3*4) ameshinfo.indexbuffershorts rawindices_array
		
		
		idstring = anodeinfo.id as string
			

		
		for aelement in ameshinfo.elements do
			(
				aname = vertex_start as string
				
				face_array = #()
				faceDir = 1
				ffacestr = 	aelement.firstface as string
				aname = idstring + "_"
				aname = aname + ffacestr

				for ri = 1 to aelement.facecount - 2 do
					(
						rawoff = aelement.firstface +  ri
						fa = rawindices_array[rawoff]
						fb = rawindices_array[rawoff+1]
						fc = rawindices_array[rawoff+2]

						if ((fa!=fb)and(fa!=fc)and(fc!=fb)) then
							(
								faceDir *= -1 ;
								if faceDir > 0 then aface = [fa,fb,fc]
								else aface = [fa,fc,fb]
								append face_array aface 
							)
					)															
			
			
				anode = mesh vertices:vertex_array faces:face_array	name:aname	

				meshop.removeillegalfaces anode 
				meshop.removedegeneratefaces anode
					
				anode.numtverts = anode.numverts
				buildtvfaces anode 
					
				for vi = 1 to anode.numverts do
				(
					settvert anode vi uvarray[vi]
					--print uvarray[vi]
				)--for vi = 1 to vertcount do
				
				-- set normals
				for vi = 1 to anode.numverts do
				(
					setnormal anode vi normal_array[vi]
					--print uvarray[vi]
				)--for vi = 1 to vertcount do				
				update anode				
						
				allfaces = #()
				for fi = 1 to anode.numfaces do					
				(
					crtface = getface anode fi
					append allfaces crtface 
				)					
					
				for fi = 1 to anode.numfaces do
				(
					settvface anode fi allfaces[fi]	
				)--for fi = 1 to facecount		
										
				allfaces = #()
				for fi = 1 to anode.numfaces do					
				(
					append allfaces fi
				)			
				meshop.unifynormals anode allfaces 	

				anodeinfo.inscene = true	
				scenenode = getnodebyname aname
				
				nodepos = anodeinfo.rawmatrix.row4
				move scenenode nodepos	
														
				converttomesh scenenode
				
				max modify mode
				
				select scenenode 
				modpanel.setcurrentobject scenenode
				subobjectlevel = 3
				meshops.removeisolatedverts scenenode
				max create mode	
			)		
	)

function getparentinfo nodeinforef allnodesref = 
	(
		for aninf in allnodesref do
		(
			if(aninf.offset == nodeinforef.parentoffset) then
			(
				return aninf
			)
		)
	)
--------------------------------------------------------------------------------------------------------------------				

filename = getopenfilename \
	types: "P3D(*.p3d)|*.p3d|All|*.*|"	

------------------------------------------GLOBALS--------------------------------------------------------------------
		nodearray = #()
---------------------------------------------------------------------------------------------------------------------
		
if filename != undefined then
	(
		f = fopen filename "rb"	
		print f	
		fseek f 0 #seek_end
		filesize = ftell f
		fseek f 8 #seek_set	
		table_offset = readlong f
		--print table_offset
		fseek f table_offset #seek_set
		object_count = readlong f
		aux = readlong f
		--format "aux  : %\n" aux 			
		aux = readlong f
		--format "aux  : %\n" aux 			
		aux = readlong f
		--format "aux  : %\n" aux 
		
		back_pointer = readlong f
		--format "back_pointer : %\n" back_pointer
		
		texture_table_offset = readlong f
		--format "texture_table_offset : %\n" texture_table_offset
		
		texture_names_off_count = readlong f
		--format "texture_names_off_count : %\n" texture_names_off_count
		
		aux = readlong f
		--format "aux  : %\n" aux
		
		obj_chunks_count = readlong f
		--format "obj_chunks_count : %\n" obj_chunks_count				

		obj_chunks_off_table_start = readlong f
		--format "obj_chunks_off_table_start : %\n" obj_chunks_off_table_start	

		aux = readlong f
		--format "aux  : %\n" aux	
		
		second_table_start = readlong f
		--format "second_table_start : %\n" second_table_start			
		
		chunks_array = #()
		fseek f obj_chunks_off_table_start #seek_set
		
		for i = 1 to obj_chunks_count do
			(				
				achunk = #()
				here = ftell f
				aoffset = readlong f
				atype = readlong f
				--format "%[%] -> offset: % , type: %\n" here i aoffset atype 
				append achunk aoffset	
				append achunk atype
				asize = 0
				append achunk asize
				append 	chunks_array achunk		
			)

		--print "first node"
		anodeinfo = readnodeinfo f 160
		anodeinfo.inscene = true
		append nodearray anodeinfo			
		--format "%\n" anodeinfo
		
		--print "second node"
		anodeinfo = readnodeinfo f 496
		anodeinfo.inscene = true			
		append nodearray anodeinfo
		--format "%\n" anodeinfo
		
		--print "nodes......."
		
		for cho in chunks_array do
			(
				if (cho[2] == 3) then
				(	
					anodeinfo = readnodeinfo f (cho[1] - 528)					
					append nodearray anodeinfo	
			
				)--if (cho[2] == 3)					
			)--for cho in chunks_array	

		nodestodraw = (nodearray.count - 2)		
		while (nodestodraw > 0) do
			(
				for ninf in nodearray do
					(
						if(ninf.inscene == false) then
							(						
								pinfo = getparentinfo ninf nodearray

								if(pinfo.inscene)then
									(
										ninf.rawmatrix.row4 = ninf.rawmatrix.row4 + pinfo.rawmatrix.row4
										--xformmat ninf.rawmatrix pinfo.rawmatrix
										drawnode f ninf
										nodestodraw -= 1
									)						
							)
					)
			)														
	)--if filename != undefined then

fclose f
print "fileclosed"		
print f
--)--for ff in files do
